/******************************************************************************
 *
 *  sis3610Lib.c  -  Driver library for operation of Struck 3610 I/O Latch
 *                  using a VxWorks 5.4 or later based Single Board computer.
 *
 *  Author: David Abbott
 *          Jefferson Lab Data Acquisition Group
 *          April 2003
 *
 *  Revision  1.0 - Initial Revision
 *
 */

#ifdef VXWORKS
#include "vxWorks.h"
#include "logLib.h"
#include "taskLib.h"
#include "intLib.h"
#include "iv.h"
#include "semLib.h"
#include "vxLib.h"
#else
#include "jvme.h"
#endif

#include "stdio.h"
#include "string.h"
#include "pthread.h"

/* Include definitions */
#include "sis3610.h"

#ifdef VXWORKS
/* Define external Functions */
IMPORT STATUS sysBusToLocalAdrs(int, char *, char **);
IMPORT STATUS intDisconnect(int);
IMPORT STATUS sysIntEnable(int);
IMPORT STATUS sysIntDisable(int);
#endif

/* Define Interrupts variables */
BOOL s3610IntRunning       = FALSE; /* running flag */
int s3610IntID             = -1; /* id number of ADC generating interrupts */
LOCAL VOIDFUNCPTR s3610IntRoutine = NULL; /* user interrupt service routine */
LOCAL int s3610IntArg      = 0; /* arg to user routine */
LOCAL int s3610IntEvCount  = 0; /* Number of Events to generate Interrupt */
LOCAL UINT32 s3610IntLevel = S3610_VME_INT_LEVEL; /* default VME interrupt level */
LOCAL UINT32 s3610IntVec   = S3610_INT_VEC; /* default interrupt Vector */
LOCAL int s3610IntMode     = S3610_INT;     /* Interupt mode */
UINT32 s3610MemOffset      = 0;

/* Define global variables */
int Ns3610 = 0; /* Number of ADCs in Crate */
volatile struct s3610_struct *s3610p[S3610_MAX_BOARDS]; /* pointers to ADC memory map */
int s3610IntCount = 0; /* Count of interrupts from ADC */
int s3610EventCount[S3610_MAX_BOARDS]; /* Count of Events taken by ADC (Event Count Register value) */
int s3610EvtReadCnt[S3610_MAX_BOARDS]; /* Count of events read from specified ADC */
int s3610RunFlag = 0;
int *s3610HistData = 0; /* Pointer to local memory for histogram data*/

#ifdef VXWORKS
SEM_ID s3610Sem; /* Semephore for Task syncronization */
#endif

/* Mutex to guard register read/writes */
pthread_mutex_t s3610Mutex = PTHREAD_MUTEX_INITIALIZER;
#define SLOCK     if(pthread_mutex_lock(&s3610Mutex)<0) perror("pthread_mutex_lock");
#define SUNLOCK   if(pthread_mutex_unlock(&s3610Mutex)<0) perror("pthread_mutex_unlock");

/**
 *  @ingroup Config
 *  @brief Initialize SIS 3610 Library. Initialize modules found through 
 *    provided routine arguments.
 *
 *   
 *  @param addr  A16 VME Address of first sis3610
 *
 *  @param addr_inc How to much increment the starting VME address to find the next
 *
 *  @param nmod Number of times to increment (number of sis3610 modules to initialize
 *
 *  @param iflag Initialization Flag (bitmask)
 *      bit  Description
 *       0   DAQ mode := Polling (if set), Interrupt (otherwise)
 *      15   No module initialization.  Just setup library pointers.
 *      
 *  @return OK if successful, otherwise ERROR.
 *
 */

STATUS
s3610Init(UINT32 addr, UINT32 addr_inc, int nmod, int iflag)
{
  int ii, res, rdata, errFlag = 0;
  int boardID = 0;
  unsigned int laddr;
  int noInit=0;
  
  /* Parse initialization flags */
  if(iflag)
    {
      if(iflag & S3610_INIT_NO_INIT)
	noInit = 1;

      if(iflag & S3610_INIT_DAQ_MODE_POLLING)
	s3610IntMode = S3610_POLL;	
    }
  else
    {
      s3610IntMode = S3610_INT;	
    }
    
  /* Check for valid address */
  if (addr == 0)
    {
      printf("%s: ERROR: Must specify a Bus (VME-based A16) address for the SIS3610\n",
	     __FUNCTION__);
      return (ERROR);
    }
  else if (addr > 0x0000ffff)
    { /* A16 Addressing */
      printf("%s: ERROR: A32/A24 Addressing not supported for the SIS 3610\n",
	     __FUNCTION__);
      return (ERROR);
    }
  else
    { /* A16 Addressing */
      if ((addr_inc == 0) || (nmod == 0))
        nmod = 1; /* assume only one ADC to initialize */

      /* get the ADC address */
#ifdef VXWORKS
      res = sysBusToLocalAdrs(0x29,(char *)addr,(char **)&laddr);
      if (res != 0)
        {
          printf("%s: ERROR in sysBusToLocalAdrs(0x29,0x%x,&laddr) \n",
		 __FUNCTION__, addr);
          return(ERROR);
        }
#else
      res = vmeBusToLocalAdrs(0x29, (char *) addr, (char **) &laddr);
      if (res != 0)
        {
          printf("%s: ERROR in vmeBusToLocalAdrs(0x29,0x%x,&laddr) \n",
		 __FUNCTION__,
		 addr);
          return (ERROR);
        }
#endif
      s3610MemOffset = laddr - addr;
    }

  Ns3610 = 0;
  for (ii = 0; ii < nmod; ii++)
    {
      s3610p[ii] = (struct s3610_struct *) (laddr + ii * addr_inc);
      /* Check if Board exists at that address */
#ifdef VXWORKS
      res = vxMemProbe((char *) &(s3610p[ii]->id),VX_READ,4,(char *)&rdata);
#else
      res = vmeMemProbe((char *) &(s3610p[ii]->id), 4, (char *) &rdata);
#endif
      if (res < 0)
        {
#ifdef VXWORKS
          printf("%s: ERROR: No addressable board at addr=0x%x\n",
		 __FUNCTION__,
		 (UINT32) s3610p[ii]);
#else
          printf("%s: ERROR: No addressable board at VME addr=0x%x\n",
		 __FUNCTION__,
		 (UINT32) s3610p[ii] - s3610MemOffset);
#endif
          s3610p[ii] = NULL;
          errFlag = 1;
          break;
        }
      else
        {
          /* Check if this is a Model 3610 */
          boardID = rdata & S3610_BOARD_ID_MASK;
          if (boardID != S3610_BOARD_ID)
            {
              printf(" ERROR: Board ID does not match: %d \n", boardID);
              return (ERROR);
            }
        }
      Ns3610++;
#ifdef VXWORKS
      printf("Initialized SIS3610 ID %d at address 0x%08x \n",ii,(UINT32) s3610p[ii]);
#else
      printf("Initialized SIS3610 ID %d at VME (USER) address 0x%x (0x%x)\n",
          ii, (UINT32) s3610p[ii] - s3610MemOffset, (UINT32) s3610p[ii]);
#endif
    }

  if(!noInit)
    {
      /* Disable/Clear all I/O Modules */
      for (ii = 0; ii < Ns3610; ii++)
	{
	  vmeWrite32(&s3610p[ii]->reset, 1);
	}
    }

  /* Initialize Interrupt variables */
  s3610IntID      = -1;
  s3610IntRunning = FALSE;
  s3610IntLevel   = S3610_VME_INT_LEVEL;
  s3610IntVec     = S3610_INT_VEC;
  s3610IntRoutine = NULL;
  s3610IntArg     = 0;
  s3610IntEvCount = 0;

  if (errFlag > 0)
    {
      printf("s3610Init: ERROR: Unable to initialize all Modules\n");

      if (Ns3610 > 0)
        printf("s3610Init: %d SIS3610(s) successfully initialized\n", Ns3610);

      return (ERROR);
    }
  else
    {
      return (OK);
    }
}

/**
 * @ingroup Status
 * @brief Print some status information of the indicated sis3610 to standard out
 *
 * @param id Library ID of sis3610
 * 
 * @param sflag Not used
 *
 */

void
s3610Status(int id, int sflag)
{

  unsigned int sreg, mid, doutput, dinput, linput;
  int iLevel, iVec, iMode, iEnable, iSource, iValid;
  int ffStatus, uoutStatus, ledStatus, latchStatus;

  if ((id < 0) || (s3610p[id] == NULL))
    {
      printf("s3610Status: ERROR : SIS3610 id %d not initialized \n", id);
      return;
    }

  /* Get info from Module */
  SLOCK;
  sreg = vmeRead32(&s3610p[id]->csr);
  mid = vmeRead32(&s3610p[id]->id);
  doutput = vmeRead32(&s3610p[id]->d_out) & 0xffff;
  dinput = vmeRead32(&s3610p[id]->d_in) & 0xffff;
  linput = vmeRead32(&s3610p[id]->l_in) & 0xffff;
  SUNLOCK;

  iLevel = (mid & S3610_INT_LEVEL_MASK) >> 8;
  iVec = (mid & S3610_INT_VEC_MASK);
  iEnable = (mid & S3610_INT_ENABLE);
  iMode = (sreg & S3610_STATUS_INT_MODE) >> 1;
  iSource = (sreg & S3610_INT_SOURCE_ENABLE_MASK) >> 20;
  iValid = (sreg & S3610_INT_SOURCE_VALID_MASK) >> 28;

  ffStatus = (sreg & S3610_FF_ENABLE_MASK) >> 16;
  uoutStatus = (sreg & S3610_USER_OUTPUT_MASK) >> 4;
  ledStatus = (sreg & S3610_STATUS_USER_LED);
  latchStatus = (sreg & S3610_LATCH_STROBE_MASK) >> 2;

#ifdef VXWORKS
  printf("\nSTATUS for SIS3610 id %d at base address 0x%x \n",id,(UINT32) s3610p[id]);
#else
  printf("\nSTATUS for SIS3610 id %d at VME (USER) base address 0x%x (0x%x)\n",
      id, (UINT32) s3610p[id] - s3610MemOffset, (UINT32) s3610p[id]);
#endif
  printf("--------------------------------------------------------------------------------\n");
  
  if (iEnable)
    {
      printf(" Interrupts Enabled  - Mode = %d  (0:RORA 1:ROAK)\n", iMode);
      printf(" VME Interrupt Level: %d   Vector: 0x%x \n", iLevel, iVec);
      printf(" Sources Enabled: 0x%x  Sources Valid: 0x%x \n", iSource, iValid);
    }
  else
    {
      printf(" Interrupts Disabled - Mode = %d  (0:RORA 1:ROAK)\n", iMode);
      printf(" VME Interrupt Level: %d   Vector: 0x%x \n", iLevel, iVec);
      printf(" Sources Enabled: 0x%x  Sources Valid: 0x%x \n", iSource, iValid);
    }
  printf("\n");
  printf("  MODULE ID   register = 0x%08x \n", mid);
  printf("\n");

  printf("  STATUS      register = 0x%08x \n", sreg);
  printf("      User Led   Enabled : %d \n", ledStatus);
  printf("      User Output Status : 0x%x \n", uoutStatus);
  printf("      Flip-Flops Enabled : 0x%x \n", ffStatus);
  printf("      Latch Strobe Status: 0x%x \n", latchStatus);

  printf("\n");
  printf("  Direct Output Register = 0x%04x \n", doutput);
  printf("  Direct Input Register  = 0x%04x \n", dinput);
  printf("  Latched Input Register = 0x%04x \n", linput);

  printf("--------------------------------------------------------------------------------\n");
  printf("\n\n");

}

/* Declare Key Address functions */

/**
 * @ingroup Config
 * @brief Perform global reset of specified sis3610
 *
 * @param id Library ID of sis3610
 * 
 */

void
s3610Reset(int id)
{
  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610Reset: ERROR : SIS3610 id %d not initialized \n", id, 0, 0,
          0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[id]->reset, 1);
  SUNLOCK;
}

/**
 * @ingroup DAQConfig
 * @brief Configure Acquisition Mode and Source
 *
 * @param id Library ID of sis3610
 * @param iLevel VME Interrupt Level
 * @param iVec   VME Interrupt Vector
 * @param iSource Which control input (1-4) to use as the trigger source
 *
 * @return Value written to Module Identification register, if
 * successful. Otherwise ERROR.
 */

int
s3610IntSet(int id, int iLevel, int iVec, int iSource)
{
  int iVal = 0;

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610IntSet: ERROR : SIS3610 id %d not initialized \n", id, 0, 0,
          0, 0, 0);
      return (ERROR);
    }

  if( (iSource < 0) || (iSource > 4) )
    {
      printf("%s: ERROR: Invalid iSource (%d)\n",
	     __FUNCTION__, iSource);
      return ERROR;
    }

  if( (iLevel < 0) || (iLevel > 7) )
    {
      printf("%s: ERROR: Invalid iLevel (%d)\n",
	     __FUNCTION__, iLevel);
      return ERROR;
    }
  
  if( (iVec < 32) || (iVec > 0xff) )
    {
      printf("%s: ERROR: Invalid iVec (%d)\n",
	     __FUNCTION__, iVec);
      return ERROR;
    }
  
  SLOCK;

  if(s3610IntMode == S3610_INT)
    {
      /* If interrupts are already enabled - don't touch anything */
      if (vmeRead32(&s3610p[id]->id) & S3610_INT_ENABLE)
	{
	  logMsg("s3610IntSet: ERROR : Interrupts are already enabled\n", 0, 0, 0,
		 0, 0, 0);
	  SUNLOCK;
	  return (ERROR);
	}
    }
  
  if (iLevel)
    iVal = (iLevel << 8);
  else
    iVal = (s3610IntLevel << 8);
  
  if (iVec)
    iVal |= ((iVec & 0xf0) + 1);
  else
    iVal |= s3610IntVec;
  
  vmeWrite32(&s3610p[id]->id, iVal);
  
  if (iSource)
    vmeWrite32(&s3610p[id]->csr, ((iSource & 0xf) << 20));

  s3610IntID = id;

  SUNLOCK;

  return (iVal);
}

/**
 * @ingroup DAQConfig
 * @brief Returns status of user outputs 1 through 4 (FlipFlops).
 *
 * @return Bit mask of user outputs (1-4) level status, if successful. Otherwise ERROR.
 */
int
s3610IntPoll(int intMask)
{
  int rval=0;
  if (s3610IntID < 0)
    {
      logMsg("s3610IntAck: ERROR : Interrupt Module ID not set\n", 
	     0, 0, 0, 0, 0, 0);
      return ERROR;
    }

  SLOCK;
  rval = (int) ((vmeRead32(&s3610p[s3610IntID]->csr) & 0xF0) >> 4);
  SUNLOCK;
  
  return rval;
}


/**
 * @ingroup DAQConfig
 * @brief Acknowledge received trigger. Allow for next.
 *
 * @param id Library ID of sis3610
 * @param intMask Bit mask of which triggers to acknowledge (bit 0 = control input 1)
 *
 */

void
s3610IntAck(int intMask)
{
  if (s3610IntID < 0)
    {
      logMsg("s3610IntAck: ERROR : Interrupt Module ID not set\n", 
	     0, 0, 0, 0, 0, 0);
      return;
    }

  /* Reset Flip-Flop */
  s3610FFReset(s3610IntID, intMask);

  if(s3610IntMode == S3610_INT)
    {
      /* Re-Enable interrupts */
      S3610_IRQ_ENABLE(s3610IntID);
    }
}

/**
 * @ingroup DAQ
 * @brief Default interrupt service routine
 *
 */
void
s3610Int(void)
{
  /* Disable interrupts */
#ifdef VXWORKS
  sysIntDisable(s3610IntLevel);
#else
  vmeBusLock();
#endif

  if(s3610IntMode == S3610_INT)
    {
      S3610_IRQ_DISABLE(s3610IntID);
    }
  
  s3610IntCount++;

  if (s3610IntRoutine != NULL)
    { /* call user routine */
      (*s3610IntRoutine)(s3610IntArg);
    }
  else
    {
      if ((s3610IntID < 0) || (s3610p[s3610IntID] == NULL))
        {
          logMsg("s3610Int: ERROR : SIS3610 id %d not initialized \n",
              s3610IntID, 0, 0, 0, 0, 0);
          return;
        }
    }

  /* Acknowledge interrupt - re-enable */
  s3610IntAck(1);

#ifdef VXWORKS
  /* Re-Enable interrupts on CPU */
  sysIntEnable(s3610IntLevel);
#else
  vmeBusUnlock();
#endif

}

/**
 * @ingroup DAQConfig
 * @brief Connect a service routine associated with an accepted trigger
 *
 * @param id Library ID of sis3610
 * @param routine Routine to Call when trigger is accepted
 * @param arg     Argument to use when calling routine
 * @param level   VME Interrupt Level
 * @param vector  VME Interrupt Vector
 *
 * @return OK if successful, otherwise ERROR.
 *
 */

STATUS
s3610IntConnect(int id, VOIDFUNCPTR routine, int arg, UINT16 level, UINT16 vector)
{
  if (s3610IntRunning)
    {
      printf("%s: ERROR : Interrupts already Initialized for SIS3610 id %d\n",
	     __FUNCTION__,
	     s3610IntID);
      return (ERROR);
    }

  s3610IntRoutine = routine;
  s3610IntArg = arg;

  /* Check for user defined VME interrupt level and vector */
  if (level == 0)
    {
      s3610IntLevel = S3610_VME_INT_LEVEL; /* use default */
    }
  else if (level > 7)
    {
      printf("%s: ERROR: Invalid VME interrupt level (%d). Must be (1-7)\n",
	     __FUNCTION__,
	     level);
      return (ERROR);
    }
  else
    {
      s3610IntLevel = level;
    }

  if (vector == 0)
    {
      s3610IntVec = S3610_INT_VEC; /* use default */
    }
  else if ((vector < 32) || (vector > 255))
    {
      printf("%s: ERROR: Invalid interrupt vector (%d). Must be (32<vector<255)\n",
	     __FUNCTION__,
	     vector);
      return (ERROR);
    }
  else
    {
      s3610IntVec = (vector & 0xf0) + 1;
    }

  /* Connect the ISR */
#ifdef VXWORKSPPC
  if((intDisconnect((int)INUM_TO_IVEC(s3610IntVec)) != 0))
    {
      printf("%s: ERROR disconnecting Interrupt\n",
	     __FUNCTION__);
      return(ERROR);
    }
#endif
#ifdef VXWORKS
  if((intConnect(INUM_TO_IVEC(s3610IntVec),s3610Int,s3610IntArg)) != 0)
    {
      printf("%s: ERROR in intConnect()\n",
	     __FUNCTION__);
      return(ERROR);
    }
#else
  if (vmeIntConnect(s3610IntVec, s3610IntLevel, s3610Int, s3610IntArg) != OK)
    {
      printf("%s: ERROR in vmeIntConnect()\n",
	     __FUNCTION__);
      return ERROR;
    }
#endif

  s3610IntSet(id, s3610IntLevel, s3610IntVec, 0);

  return (OK);
}

/**
 * @ingroup DAQConfig
 * @brief Enable triggers to be accepted
 *
 * @param intMask Bit mask of control inputs to use for accepted triggers
 */

void
s3610IntEnable(int intMask)
{

  if (s3610IntID < 0)
    {
      logMsg("s3610IntEnable: ERROR : Interrupt Module ID not set\n", 0, 0, 0,
          0, 0, 0);
      return;
    }

  if (s3610IntRunning)
    {
      logMsg(
          "s3610IntEnable: ERROR : Interrupts already Enabled for SIS3610 id %d\n",
          s3610IntID, 0, 0, 0, 0, 0);
      return;
    }

#ifdef VXWORKS
  sysIntDisable(s3610IntLevel);
#endif

  SLOCK;
  if(s3610IntMode == S3610_INT)
    {
      if (intMask == 0)
	{ /* Check if any sources are enabled */
	  if ((vmeRead32(&s3610p[s3610IntID]->csr) & S3610_INT_SOURCE_ENABLE_MASK)
	      == 0)
	    {
          logMsg("s3610IntEnable: ERROR : No Interrupt sources are specified\n",
		 0, 0, 0, 0, 0, 0);
          SUNLOCK;
          return;
	    }
	}
      else
	{
	  if(s3610IntMode == S3610_INT)
	    {
	      /* Enable Interrupts */
	      vmeWrite32(&s3610p[s3610IntID]->csr, (intMask & 0xf) << 20);
	    }
	}
    }

  SUNLOCK;
  s3610FFEnable(s3610IntID, (intMask & 0xf));

  if(s3610IntMode == S3610_INT)
    {
      S3610_IRQ_ENABLE(s3610IntID);
#ifdef VXWORKS
      sysIntEnable(s3610IntLevel);
#endif
    }
  
  s3610IntRunning = 1;


  return;
}

/**
 * @ingroup DAQConfig
 * @brief Enable RORA (Release On Register Access) interrupter style
 *
 */

void
s3610RORA()
{

  if (s3610IntID < 0)
    {
      logMsg("s3610RORA: ERROR : Interrupt Module ID not set\n", 0, 0, 0, 0, 0,
          0);
      return;
    }

  if (s3610IntRunning)
    {
      logMsg("s3610RORA: ERROR : Interrupts Enabled for SIS3610 id %d\n",
          s3610IntID, 0, 0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[s3610IntID]->csr, S3610_CNTL_SET_RORA);
  SUNLOCK;

  return;
}

/**
 * @ingroup DAQConfig
 * @brief Enable RORA (Release On Acknowledge) interrupter style
 *
 */

void
s3610ROAK()
{

  if (s3610IntID < 0)
    {
      logMsg("s3610ROAK: ERROR : Interrupt Module ID not set\n", 0, 0, 0, 0, 0,
          0);
      return;
    }

  if (s3610IntRunning)
    {
      logMsg("s3610ROAK: ERROR : Interrupts Enabled for SIS3610 id %d\n",
          s3610IntID, 0, 0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[s3610IntID]->csr, S3610_CNTL_SET_ROAK);
  SUNLOCK;

  return;
}

/**
 * @ingroup DAQConfig
 * @brief Disable acceptance of triggers
 *
 * @param intMask Bit mask of controller inputs to disable.
 *
 */

void
s3610IntDisable(int intMask)
{

  if (s3610IntID < 0)
    {
      logMsg("s3610IntDisable: ERROR : Interrupt Module ID not set\n", 0, 0, 0,
          0, 0, 0);
      return;
    }


  if(s3610IntMode == S3610_INT)
    {
#ifdef VXWORKS
      sysIntDisable(s3610IntLevel);
#endif
      S3610_IRQ_DISABLE(s3610IntID);
    }
  
  SLOCK;
  s3610IntRunning = 0;

  if (intMask)
    {
      if(s3610IntMode == S3610_INT)
	{
	  vmeWrite32(&s3610p[s3610IntID]->csr, (intMask & 0xf) << 28);
	}

      SUNLOCK;
      s3610FFDisable(s3610IntID, (intMask & 0xf));
    }
  else
    {
      SUNLOCK;
    }

  return;

}

/**
 * @ingroup Config
 * @brief Read or Write a value to the CSR (offset 0x0) if specified sis3610
 *
 * @param id Library ID of sis3610
 * @param val Value to write to register (must be > 0)
 *
 * @return Value stored at CSR if successful, otherwise 0.
 *
 */

unsigned int
s3610CSR(int id, unsigned int val)
{
  unsigned int rval;

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610CSR: ERROR : SIS3610 id %d not initialized \n", id, 0, 0, 0,
          0, 0);
      return (0);
    }

  SLOCK;
  if (val)
    vmeWrite32(&s3610p[id]->csr, val);

  rval = vmeRead32(&s3610p[id]->csr);
  SUNLOCK;

  return (rval);

}

/**
 * @ingroup Config
 * @brief Turn on the Front Panel LED
 * @param id Library ID of sis3610
 *
 */

void
s3610SetLED(int id)
{
  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610SetLED: ERROR : SIS3610 id %d not initialized \n", id, 0, 0,
          0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[id]->csr, S3610_CNTL_SET_USER_LED);
  SUNLOCK;

  return;
}

/**
 * @ingroup Config
 * @brief Turn off the Front Panel LED
 * @param id Library ID of sis3610
 *
 */

void
s3610ClearLED(int id)
{
  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610ClearLED: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[id]->csr, S3610_CNTL_CLR_USER_LED);
  SUNLOCK;

  return;
}

/**
 * @ingroup Config
 * @brief Enable Flip-flop of specified control inputs
 *
 * @param id Library ID of sis3610
 * @param val Bit mask of control inputs to enable
 *
 */

void
s3610FFEnable(int id, int val)
{

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610FFEnable: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return;
    }

  if (val == 0) /* Enable control input 1 in FF mode only */
    val = 1;

  SLOCK;
  vmeWrite32(&s3610p[id]->csr, ((val & 0xf) << 16));
  SUNLOCK;

  return;
}

/**
 * @ingroup Config
 * @brief Disable Flip-flop of specified control inputs
 *
 * @param id Library ID of sis3610
 * @param val Bit mask of control inputs to disable
 *
 */

void
s3610FFDisable(int id, int val)
{

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610FFDisable: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return;
    }

  if (val == 0) /* Disable control input 1 FF mode only */
    val = 1;

  SLOCK;
  vmeWrite32(&s3610p[id]->csr, ((val & 0xf) << 24));
  SUNLOCK;

  return;
}

/**
 * @ingroup Config
 * @brief Reset (set to 0) Flip-flop of specified control inputs
 *
 * @param id Library ID of sis3610
 * @param val Bit mask of control inputs to disable
 *
 */

void
s3610FFReset(int id, int val)
{

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610FFReset: ERROR : SIS3610 id %d not initialized \n", id, 0, 0,
          0, 0, 0);
      return;
    }

  if (val == 0) /* Reset control input 1 FF mode only */
    val = 1;

  SLOCK;
  vmeWrite32(&s3610p[id]->csr, ((val & 0xf) << 12));
  SUNLOCK;

  return;
}

/**
 * @ingroup Status
 * @brief Return the status of the 16 output data levels
 *
 * @param id Library ID of sis3610
 *
 * @return Bit mask (16bits) of output data levels, if successful.  Otherwise ERROR.
 */

int
s3610ReadOutput(int id)
{
  int res = 0;

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610ReadOutput: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return (ERROR);
    }

  SLOCK;
  res = vmeRead32(&s3610p[id]->d_out) & 0xffff;
  SUNLOCK;

  return (res);
}

/**
 * @ingroup Status
 * @brief Return the status of the 16 input data levels
 *
 * @param id Library ID of sis3610
 *
 * @return Bit mask (16bits) of input data levels, if successful.  Otherwise ERROR.
 */

int
s3610ReadInput(int id)
{
  int res = 0;

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610ReadInput: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return (ERROR);
    }

  SLOCK;
  res = vmeRead32(&s3610p[id]->d_in) & 0xffff;
  SUNLOCK;

  return (res);
}

/**
 * @ingroup Status
 * @brief Return the status of the 16 latched input data levels
 *
 * @param id Library ID of sis3610
 *
 * @return Bit mask (16bits) of latched input data levels, if successful.  Otherwise ERROR.
 */

int
s3610ReadLatch(int id)
{
  int res = 0;

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610ReadLatch: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return (ERROR);
    }

  SLOCK;
  res = vmeRead32(&s3610p[id]->l_in) & 0xffff;
  SUNLOCK;

  return (res);
}

/**
 * @ingroup Config
 * @brief Set the bit pattern of the output data levels
 *
 * @param id Library ID of sis3610
 * @param val Bit mask (16bits) of output data levels to set and unset
 *
 */

void
s3610WriteOutput(int id, unsigned int val)
{

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610WriteOutput: ERROR : SIS3610 id %d not initialized \n", id,
          0, 0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[id]->d_out, val);
  SUNLOCK;

  return;
}

/**
 * @ingroup Config
 * @brief Set the bit pattern of the output data levels.  The specified pattern is OR'd with the current output data level.
 *
 * @param id Library ID of sis3610
 * @param val Bit mask (16bits) of output data levels to set
 *
 */

void
s3610WriteBits(int id, unsigned int val)
{

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610WriteBits: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[id]->jk_out, val);
  SUNLOCK;

  return;
}

/**
 * @ingroup Config
 * @brief Clear the bit pattern of the output data levels.
 *
 * @param id Library ID of sis3610
 * @param val Bit mask (16bits) of output data levels to clear
 *
 */

void
s3610ClearBits(int id, unsigned int val)
{

  if ((id < 0) || (s3610p[id] == NULL))
    {
      logMsg("s3610ClearBits: ERROR : SIS3610 id %d not initialized \n", id, 0,
          0, 0, 0, 0);
      return;
    }

  if (val == 0)
    val = 0xffff;

  SLOCK;
  vmeWrite32(&s3610p[id]->jk_out, (val << 16));
  SUNLOCK;

  return;
}

void
s3610IrqEnable(int intID)
{
  if (s3610p[intID] == NULL)
    {
      logMsg("s3610ClearBits: ERROR : SIS3610 id %d not initialized \n", intID, 0,
          0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[intID]->id,					
	       vmeRead32(&s3610p[intID]->id) | S3610_INT_ENABLE);
  SUNLOCK;
}

void
s3610IrqDisable(int intID)
{
  if (s3610p[intID] == NULL)
    {
      logMsg("s3610ClearBits: ERROR : SIS3610 id %d not initialized \n", intID, 0,
          0, 0, 0, 0);
      return;
    }

  SLOCK;
  vmeWrite32(&s3610p[intID]->id,					
	       vmeRead32(&s3610p[intID]->id) & ~S3610_INT_ENABLE);
  SUNLOCK;
}
